import java.util.LinkedList;
import java.util.List;
import java.util.ArrayList;
import java.awt.Graphics2D;

public class Pts {

  interface Function {
    double call(double arg);
  }

  static class Functions {
    static Function literal(final double val) {
      return new Function() {
        @Override
        public double call(double ignore) {
          return val;
        }
      };
    }

    static Function multiply(final Function lhs, final Function rhs) {
      return new Function() {
        @Override
        public double call(double x) {
          return lhs.call(x) * rhs.call(x);
        }
      };
    }

    static Function add(final Function lhs, final Function rhs) {
      return new Function() {
        @Override
        public double call(double x) {
          return lhs.call(x) + rhs.call(x);
        }
      };
    }
  }

  static <T> List<T> toList(T[] array) {
    List<T> list = new LinkedList<T>();
    for (T val : array) {
      list.add(val);
    }
    return list;
  }

  static class Point {
    private final double values[];

    Point(double x, double y, double z) {
      this.values = new double[] {x, y, z, 1.0};
    }

    double get(int i) {
      return values[i];
    }

    void set(double val, int i) {
      values[i] = val;
    }

    @Override
    public String toString() {
      return "{ " + values[0] + ", " + values[1] + ", " + values[2] + "}";
    }
  }

  static class Matrix {
    private static final int N = 4; // 4x4 matrix

    private final ArrayList<Function> values;

    Matrix(Function... newValues) {
      this(toList(newValues));
    }

    Matrix() {
      this.values = new ArrayList<Function>(N * N);
      for (int i = 0; i < N; i++) {
        for (int j = 0; j < N; j++) {
          values.add(Functions.literal(i == j ? 1 : 0));
        }
      }
    }

    Matrix(List<Function> newValues) {
      if (newValues.size() == 16) {
        values = new ArrayList<Function>(newValues);
      } else if (newValues.size() == 9) {
        // Extend to 4x4 matrix

        // Row 1
        values = new ArrayList<Function>(newValues.subList(0, 3));
        values.add(Functions.literal(0));

        // Row 2
        values.addAll(newValues.subList(3, 6));
        values.add(Functions.literal(0));

        // Row 3
        values.addAll(newValues.subList(6, 9));
        values.add(Functions.literal(0));

        // Row 4
        values.add(Functions.literal(0));
        values.add(Functions.literal(0));
        values.add(Functions.literal(0));
        values.add(Functions.literal(1));
      } else {
        throw new AssertionError("Unsupported array size: " + newValues.size());
      }
    }

    Function get(int col, int row) {
      return values.get(4 * row + col);
    }

    ArrayList<Function> getRow(int row) {
      return new ArrayList<Function>(values.subList(4 * row, 4 * row + N));
    }

    ArrayList<Function> getCol(int col) {
      ArrayList<Function> result = new ArrayList<Function>();
      result.add(get(col, 0));
      result.add(get(col, 1));
      result.add(get(col, 2));
      result.add(get(col, 3));
      return result;
    }

    void set(int col, int row, Function value) {
      values.set(4 * row + col, value);
    }

    static Matrix multiply(Matrix lhs, Matrix rhs) {
      Matrix result = new Matrix();
      for (int col = 0; col < N; col++) {
        for (int row = 0; row < N; row++) {
          ArrayList<Function> lValues = lhs.getRow(row);
          ArrayList<Function> rValues = rhs.getCol(col);
          Function total = Functions.literal(0);
          for (int i = 0; i < N; i++) {
            total = Functions.add(total,
                Functions.multiply(lValues.get(i), rValues.get(i)));
          }
          result.set(col, row, total);
        }
      }
      return result;
    }

    static Point multiply(Matrix m, Point point, double t) {
      Point result = new Point(0, 0, 0);
      for (int row = 0; row < N; row++) {
        ArrayList<Function> values = m.getRow(row);
        double total = 0.0;
        for (int i = 0; i < N; i++) {
          total += values.get(i).call(t) * point.get(i);
        }
        result.set(total, row);
      }
      return result;
    }

    @Override
    public String toString() {
      StringBuilder b = new StringBuilder("[\n");
      for (int i = 0; i < N; i++) {
        for (int j = 0; j < N; j++) {
          b.append(get(j, i).call(1)).append(",  ");
        }
        b.append("\n");
      }
      b.append("]");
      return b.toString();
    }
  }

  static void render(Matrix transform, List<Point> points, double t) {
    System.out.println("Rendering t=" + t);
    for (Point p : points) {
      System.out.println(Matrix.multiply(transform, p, t));
    }
  }

  public static void main(String[] args) {
    try {
      run();
    } catch (InterruptedException e) {
      System.out.println("Interrupted");
    }
  }

  private static void run() throws InterruptedException {
