open Ast
open Printf

module StringMap = Map.Make(String)

type id_type = IdConst | IdPoint | IdPoints | IdMatrix | IdFunc

let add_id_type map def = match def with
    ConstDef(id, expr) -> StringMap.add id IdConst map
  | PointDef(id, point) -> StringMap.add id IdPoint map
  | PointsDef(id, points) -> StringMap.add id IdPoints map
  | MatrixDef(id, m) -> StringMap.add id IdMatrix map
  | FuncDef(id, arg, expr) -> StringMap.add id IdFunc map

let rec build_id_type_map stmts map = match stmts with
    [] -> map
  | hd :: tl ->
      (match hd with
          Def(def) ->
            let id_map = build_id_type_map tl map in
            add_id_type id_map def
        | _ -> map)

let rec translate_expr = function
    Lit(x) -> x
  | Binop(e1, op, e2) ->
      let v1 = translate_expr e1 and v2 = translate_expr e2 in
      (match op with
          Add -> "(" ^ v1 ^ " + " ^ v2 ^ ")"
        | Sub -> "(" ^ v1 ^ " - " ^ v2 ^ ")"
        | Mul -> "(" ^ v1 ^ " * " ^ v2 ^ ")"
        | Div -> "(" ^ v1 ^ " / " ^ v2 ^ ")"
        | Pow -> "Math.pow(" ^ v1 ^ ", " ^ v2 ^ ")")
   | Id(x) -> x
   | FuncApplication(id, arg) -> id ^ "(" ^ (translate_expr arg) ^ ")"

let translate_point = function
    PointLit(p) -> "new Point(" ^ (translate_expr p.x) ^
        ", " ^ (translate_expr p.y) ^ ", " ^ (translate_expr p.z) ^ ")"
  | PointId(pid) -> pid

let rec add_points list_id points = match points with
    [] -> ""
  | hd :: tl -> ";\n " ^ list_id ^ ".add(" ^ (translate_point hd) ^ ")" ^ (add_points list_id tl)

let translate_function_literal expr = "Functions.literal(" ^ (translate_expr expr) ^ ")"

let translate_matrix_expr expr id_map = match expr with
    Id(name) -> let id_type = StringMap.find name id_map in
        (match id_type with
            IdFunc -> name
          | _ ->  translate_function_literal expr)
  | _ -> translate_function_literal expr

let rec translate_matrix mlist id_map = match mlist with
    [] -> ""
  | [x] -> (translate_matrix_expr x id_map)
  | hd :: tl -> (translate_matrix_expr hd id_map) ^ "," ^ (translate_matrix tl id_map)

let rec translate_matrix_chain = function
    [] -> ""
  | [x] -> x
  | hd :: tl -> "Matrix.multiply(" ^ hd ^ ", " ^ (translate_matrix_chain tl) ^ ")"

let translate_def def id_map = match def with
    ConstDef(id, expr) -> "final double " ^ id ^ " = " ^ (translate_expr expr)
  | PointDef(id, point) -> "final Point " ^ id ^ " = " ^ (translate_point point)
  | PointsDef(id, points) -> "final LinkedList<Point> " ^ id ^ " = new LinkedList<Point>()" ^
      (add_points id points)
  | MatrixDef(id, m) ->
      let var_name = "final Matrix " ^ id ^ " = " in
      (match m with
          MatrixLit(values) ->  var_name ^ "new Matrix(" ^ (translate_matrix values id_map) ^ ")"
        | MatrixChain(ids) -> var_name ^ (translate_matrix_chain ids))
  | FuncDef(id, arg, expr) ->
      "final Function " ^ id ^ " = new Function() {\n" ^
      "  @Override public double call(double " ^ arg ^") {\n" ^
      "    return " ^ (translate_expr expr) ^ ";\n" ^
      "  }\n" ^
      "}"

let rec translate_prog stmt_list id_map = match stmt_list with
    [] -> ""
  | [x] -> (translate_stmt x id_map) ^ ";\n"
  | hd :: tl -> (translate_stmt hd id_map) ^ ";\n" ^ (translate_prog tl id_map)
and translate_stmt stmt id_map = match stmt with
    Def(def) -> translate_def def id_map
  | Sleep(time) -> "Thread.sleep(" ^ (translate_expr time) ^ ")"
  | Render(m, p, t) -> "render(" ^ m ^ ", " ^ p ^ ", " ^ (translate_expr t) ^ ")"
  | AddPoint(pts, p) -> pts ^ ".add(" ^ (translate_point p) ^ ")"
  | ForLoop(i, low, high, body) ->
      "for (int " ^ i ^ " = " ^ low ^ "; " ^ i ^ " <= " ^ high ^ "; ++" ^ i ^ ") {\n" ^
      (translate_prog body id_map) ^ "}"



let _ =
  let lexbuf = Lexing.from_channel stdin in
  let program = Parser.program Scanner.token lexbuf in
  let id_map = build_id_type_map program StringMap.empty in
  let file = open_out "pts.output" in
  fprintf file "%s\n" (translate_prog program id_map);
  close_out file;
